@@ -194,7 +194,7 @@ int apn_start(struct apn_ctx *apn)
 		LOGPAPN(LOGL_INFO, apn, "Opened TUN device %s\n", apn->tun.tun->devname);
 
 		/* Register with libosmcoore */
-		osmo_fd_setup(&apn->tun.fd, apn->tun.tun->fd, BSC_FD_READ, ggsn_tun_fd_cb, apn, 0);
+		osmo_fd_setup(&apn->tun.fd, apn->tun.tun->fd, OSMO_FD_READ, ggsn_tun_fd_cb, apn, 0);
 		osmo_fd_register(&apn->tun.fd);
 
 		/* Set TUN library callback */
@@ -467,9 +467,13 @@ int create_context_ind(struct pdp_t *pdp)
 		return 0;
 	}
 
-	/* FIXME: we manually force all context requests to dynamic here! */
-	if (pdp->eua.l > 2)
-		pdp->eua.l = 2;
+	/* FIXME: implement context request for static IP addresses! */
+	if (pdp->eua.l > 2) {
+		LOGPPDP(LOGL_ERROR, pdp, "Static IP addresses not supported: %s\n",
+			osmo_hexdump(pdp->eua.v, pdp->eua.l));
+		gtp_create_context_resp(gsn, pdp, GTPCAUSE_NOT_SUPPORTED);
+		return 0;
+	}
 
 	memcpy(pdp->qos_neg0, pdp->qos_req0, sizeof(pdp->qos_req0));
 
@@ -589,7 +593,7 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 	struct iphdr *iph = (struct iphdr *)pack;
 	struct ip6_hdr *ip6h = (struct ip6_hdr *)pack;
 	struct ippool_t *pool;
-	char straddr[INET6_ADDRSTRLEN];
+	char straddr[2][INET6_ADDRSTRLEN];
 	uint8_t pref_offset;
 
 	switch (iph->version) {
@@ -622,17 +626,41 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 		return 0;
 
 	if (ippool_getip(pool, &ipm, &dst)) {
-		LOGTUN(LOGL_DEBUG, tun, "Received packet for APN(%s) with no PDP contex! (%s)\n",
-			apn->cfg.name,
-			iph->version == 4 ?
-			inet_ntop(AF_INET, &iph->saddr, straddr, sizeof(straddr)) :
-			inet_ntop(AF_INET6, &ip6h->ip6_src, straddr, sizeof(straddr)));
+		LOGTUN(LOGL_DEBUG, tun, "APN(%s) Rx DL data packet for IP address outside "
+		       "pool of managed addresses: %s <- %s\n",
+		       apn->cfg.name,
+		       iph->version == 4 ?
+		         inet_ntop(AF_INET, &iph->daddr, straddr[0], sizeof(straddr[0])) :
+		         inet_ntop(AF_INET6, &ip6h->ip6_dst, straddr[0], sizeof(straddr[0])),
+		       iph->version == 4 ?
+		         inet_ntop(AF_INET, &iph->saddr, straddr[1], sizeof(straddr[1])) :
+		         inet_ntop(AF_INET6, &ip6h->ip6_src, straddr[1], sizeof(straddr[1])));
 		return 0;
 	}
-	LOGTUN(LOGL_DEBUG, tun, "Received packet for APN(%s)\n", apn->cfg.name);
 
-	if (ipm->peer)		/* Check if a peer protocol is defined */
-		gtp_data_req(apn->ggsn->gsn, (struct pdp_t *)ipm->peer, pack, len);
+	if (ipm->peer)	{	/* Check if a peer protocol is defined */
+		struct pdp_t *pdp = (struct pdp_t *)ipm->peer;
+		LOGTUN(LOGL_DEBUG, tun, "APN(%s) Rx DL data packet for PDP(%s:%u): %s <- %s\n",
+		       apn->cfg.name,
+		       imsi_gtp2str(&(pdp)->imsi), (pdp)->nsapi,
+		       iph->version == 4 ?
+		         inet_ntop(AF_INET, &iph->daddr, straddr[0], sizeof(straddr[0])) :
+		         inet_ntop(AF_INET6, &ip6h->ip6_dst, straddr[0], sizeof(straddr[0])),
+		       iph->version == 4 ?
+		         inet_ntop(AF_INET, &iph->saddr, straddr[1], sizeof(straddr[1])) :
+		         inet_ntop(AF_INET6, &ip6h->ip6_src, straddr[1], sizeof(straddr[1])));
+		gtp_data_req(apn->ggsn->gsn, pdp, pack, len);
+	} else {
+		LOGTUN(LOGL_DEBUG, tun, "APN(%s) Rx DL data packet for IP address with no "
+		       "associated PDP Ctx: %s <- %s\n",
+		       apn->cfg.name,
+		       iph->version == 4 ?
+ 		         inet_ntop(AF_INET, &iph->daddr, straddr[0], sizeof(straddr[0])) :
+ 		         inet_ntop(AF_INET6, &ip6h->ip6_dst, straddr[0], sizeof(straddr[0])),
+ 		       iph->version == 4 ?
+ 		         inet_ntop(AF_INET, &iph->saddr, straddr[1], sizeof(straddr[1])) :
+ 		         inet_ntop(AF_INET6, &ip6h->ip6_src, straddr[1], sizeof(straddr[1])));
+	}
 	return 0;
 }
 
@@ -704,7 +732,7 @@ static int ggsn_tun_fd_cb(struct osmo_fd *fd, unsigned int what)
 {
 	struct apn_ctx *apn = fd->data;
 
-	OSMO_ASSERT(what & BSC_FD_READ);
+	OSMO_ASSERT(what & OSMO_FD_READ);
 
 	return tun_decaps(apn->tun.tun);
 }
@@ -715,7 +743,7 @@ static int ggsn_gtp_fd_cb(struct osmo_fd *fd, unsigned int what)
 	struct ggsn_ctx *ggsn = fd->data;
 	int rc;
 
-	OSMO_ASSERT(what & BSC_FD_READ);
+	OSMO_ASSERT(what & OSMO_FD_READ);
 
 	switch (fd->priv_nr) {
 	case 0:
@@ -809,15 +837,15 @@ int ggsn_start(struct ggsn_ctx *ggsn)
 		ggsn->gsn->gsnu = ggsn->cfg.gtpu_addr.v4;
 
 	/* Register File Descriptors */
-	osmo_fd_setup(&ggsn->gtp_fd0, ggsn->gsn->fd0, BSC_FD_READ, ggsn_gtp_fd_cb, ggsn, 0);
+	osmo_fd_setup(&ggsn->gtp_fd0, ggsn->gsn->fd0, OSMO_FD_READ, ggsn_gtp_fd_cb, ggsn, 0);
 	rc = osmo_fd_register(&ggsn->gtp_fd0);
 	OSMO_ASSERT(rc == 0);
 
-	osmo_fd_setup(&ggsn->gtp_fd1c, ggsn->gsn->fd1c, BSC_FD_READ, ggsn_gtp_fd_cb, ggsn, 1);
+	osmo_fd_setup(&ggsn->gtp_fd1c, ggsn->gsn->fd1c, OSMO_FD_READ, ggsn_gtp_fd_cb, ggsn, 1);
 	rc = osmo_fd_register(&ggsn->gtp_fd1c);
 	OSMO_ASSERT(rc == 0);
 
-	osmo_fd_setup(&ggsn->gtp_fd1u, ggsn->gsn->fd1u, BSC_FD_READ, ggsn_gtp_fd_cb, ggsn, 2);
+	osmo_fd_setup(&ggsn->gtp_fd1u, ggsn->gsn->fd1u, OSMO_FD_READ, ggsn_gtp_fd_cb, ggsn, 2);
 	rc = osmo_fd_register(&ggsn->gtp_fd1u);
 	OSMO_ASSERT(rc == 0);
 
