@@ -389,10 +389,9 @@ DEFUN(cfg_ggsn_no_echo_interval, cfg_ggsn_no_echo_interval_cmd,
 	NO_STR "Send an echo request to this static GGSN every interval.\n")
 {
 	struct ggsn_ctx *ggsn = (struct ggsn_ctx *) vty->index;
-	int prev_interval = ggsn->cfg.echo_interval;
 	struct sgsn_peer *sgsn;
 
-	if (prev_interval == ggsn->cfg.echo_interval)
+	if (ggsn->cfg.echo_interval == 0)
 		return CMD_SUCCESS;
 
 	ggsn->cfg.echo_interval = 0;
@@ -533,9 +532,11 @@ DEFUN(cfg_apn_ip_prefix, cfg_apn_ip_prefix_cmd,
 	struct in46_prefix *pfx;
 
 	/* first update our parsed prefix */
-	if (!strcmp(argv[0], "static"))
+	if (!strcmp(argv[0], "static")) {
 		pfx = &apn->v4.cfg.static_prefix;
-	else
+		vty_out(vty, "%% static IP addresses currently not yet supported%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	} else
 		pfx = &apn->v4.cfg.dynamic_prefix;
 	str2prefix(pfx, argv[1]);
 
@@ -567,9 +568,11 @@ DEFUN(cfg_apn_ipv6_prefix, cfg_apn_ipv6_prefix_cmd,
 	struct apn_ctx *apn = (struct apn_ctx *) vty->index;
 	struct in46_prefix *pfx;
 
-	if (!strcmp(argv[0], "static"))
+	if (!strcmp(argv[0], "static")) {
 		pfx = &apn->v6.cfg.static_prefix;
-	else
+		vty_out(vty, "%% static IP addresses currently not yet supported%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	} else
 		pfx = &apn->v6.cfg.dynamic_prefix;
 	str2prefix(pfx, argv[1]);
 	return CMD_SUCCESS;
@@ -699,6 +702,10 @@ DEFUN(cfg_apn_no_shutdown, cfg_apn_no_shutdown_cmd,
 	struct apn_ctx *apn = (struct apn_ctx *) vty->index;
 
 	if (apn->cfg.shutdown) {
+		if (!apn->tun.cfg.dev_name) {
+			vty_out(vty, "%% Failed to start APN, tun-device is not configured%s", VTY_NEWLINE);
+			return CMD_WARNING;
+		}
 		if (apn_start(apn) < 0) {
 			vty_out(vty, "%% Failed to start APN, check log for details%s", VTY_NEWLINE);
 			return CMD_WARNING;
@@ -894,11 +901,16 @@ DEFUN(show_pdpctx_imsi, show_pdpctx_imsi_cmd,
 		return CMD_WARNING;
 	}
 
-	imsi = strtoull(argv[1], NULL, 10);
+	if (strlen(argv[1]) < 6 || strlen(argv[1]) > 15) {
+		vty_out(vty, "%% Invalid IMSI '%s'%s", argv[1], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	imsi = gtp_imsi_str2gtp(argv[1]);
 
 	if (argc > 2) {
 		nsapi = atoi(argv[2]);
-		if (gtp_pdp_getimsi(ggsn->gsn, &pdp, imsi, nsapi)) {
+		if (!gtp_pdp_getimsi(ggsn->gsn, &pdp, imsi, nsapi)) {
 			show_one_pdp(vty, pdp);
 			num_found++;
 		}
