@@ -31,14 +31,7 @@
 #include "pdp.h"
 #include "gtp.h"
 #include "lookupa.h"
-
-/* ***********************************************************
- * Global variables TODO: most should be moved to gsn_t
- *************************************************************/
-
-static struct pdp_t pdpa[PDP_MAX];	/* PDP storage */
-static struct pdp_t *hashtid[PDP_MAX];	/* Hash table for IMSI + NSAPI */
-/* struct pdp_t* haship[PDP_MAX];  Hash table for IP and network interface */
+#include "queue.h"
 
 /* ***********************************************************
  * Functions related to PDP storage
@@ -112,11 +105,16 @@ static struct pdp_t *hashtid[PDP_MAX];	/* Hash table for IMSI + NSAPI */
  *
  *************************************************************/
 
-int pdp_init()
+static struct gsn_t *g_gsn;
+
+int pdp_init(struct gsn_t *gsn)
 {
-	memset(&pdpa, 0, sizeof(pdpa));
-	memset(&hashtid, 0, sizeof(hashtid));
-	/*  memset(&haship, 0, sizeof(haship)); */
+	if (!g_gsn) {
+		g_gsn = gsn;
+	} else {
+		LOGP(DLGTP, LOGL_FATAL, "This interface is depreacted and doesn't support multiple GGSN!");
+		return -1;
+	}
 
 	return 0;
 }
@@ -124,6 +122,13 @@ int pdp_init()
 int pdp_newpdp(struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi,
 	       struct pdp_t *pdp_old)
 {
+	return gtp_pdp_newpdp(g_gsn, pdp, imsi, nsapi, pdp_old);
+}
+
+int gtp_pdp_newpdp(struct gsn_t *gsn, struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi,
+	       struct pdp_t *pdp_old)
+{
+	struct pdp_t *pdpa = gsn->pdpa;
 	int n;
 	for (n = 0; n < PDP_MAX; n++) {	/* TODO: Need to do better than linear search */
 		if (pdpa[n].inuse == 0) {
@@ -133,6 +138,7 @@ int pdp_newpdp(struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi,
 			else
 				memset(*pdp, 0, sizeof(struct pdp_t));
 			(*pdp)->inuse = 1;
+			(*pdp)->gsn = gsn;
 			(*pdp)->imsi = imsi;
 			(*pdp)->nsapi = nsapi;
 			(*pdp)->fllc = (uint16_t) n + 1;
@@ -151,7 +157,7 @@ int pdp_newpdp(struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi,
 			}
 			/* Default: Generate G-PDU sequence numbers on Tx */
 			(*pdp)->tx_gpdu_seq = true;
-
+			INIT_LLIST_HEAD(&(*pdp)->qmsg_list_req);
 			return 0;
 		}
 	}
@@ -160,6 +166,18 @@ int pdp_newpdp(struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi,
 
 int pdp_freepdp(struct pdp_t *pdp)
 {
+	struct qmsg_t *qmsg, *qmsg2;
+	struct pdp_t *pdpa = pdp->gsn->pdpa;
+	int rc;
+
+	/* Remove all enqueued messages belonging to this pdp from req tx transmit
+	   queue. queue_freemsg will call llist_del(). */
+	llist_for_each_entry_safe(qmsg, qmsg2, &pdp->qmsg_list_req, entry) {
+		if ((rc = queue_freemsg(pdp->gsn->queue_req, qmsg)))
+			LOGP(DLGTP, LOGL_ERROR,
+			     "Failed freeing qmsg from qmsg_list_req during pdp_freepdp()! %d\n", rc);
+	}
+
 	pdp_tiddel(pdp);
 
 	/* Remove any references in primary context */
@@ -174,12 +192,20 @@ int pdp_freepdp(struct pdp_t *pdp)
 
 int pdp_getpdp(struct pdp_t **pdp)
 {
-	*pdp = &pdpa[0];
+	*pdp = &g_gsn->pdpa[0];
 	return 0;
 }
 
 int pdp_getgtp0(struct pdp_t **pdp, uint16_t fl)
 {
+	return gtp_pdp_getgtp0(g_gsn, pdp, fl);
+}
+
+
+int gtp_pdp_getgtp0(struct gsn_t *gsn, struct pdp_t **pdp, uint16_t fl)
+{
+	struct pdp_t *pdpa = gsn->pdpa;
+
 	if ((fl > PDP_MAX) || (fl < 1)) {
 		return EOF;	/* Not found */
 	} else {
@@ -194,6 +220,13 @@ int pdp_getgtp0(struct pdp_t **pdp, uint16_t fl)
 
 int pdp_getgtp1(struct pdp_t **pdp, uint32_t tei)
 {
+	return gtp_pdp_getgtp1(g_gsn, pdp, tei);
+}
+
+int gtp_pdp_getgtp1(struct gsn_t *gsn, struct pdp_t **pdp, uint32_t tei)
+{
+	struct pdp_t *pdpa = gsn->pdpa;
+
 	if ((tei > PDP_MAX) || (tei < 1)) {
 		return EOF;	/* Not found */
 	} else {
@@ -209,6 +242,12 @@ int pdp_getgtp1(struct pdp_t **pdp, uint32_t tei)
 /* get a PDP based on the *peer* address + TEI-Data.  Used for matching inbound Error Ind */
 int pdp_getgtp1_peer_d(struct pdp_t **pdp, const struct sockaddr_in *peer, uint32_t teid_gn)
 {
+	return gtp_pdp_getgtp1_peer_d(g_gsn, pdp, peer, teid_gn);
+}
+
+int gtp_pdp_getgtp1_peer_d(struct gsn_t *gsn, struct pdp_t **pdp, const struct sockaddr_in *peer, uint32_t teid_gn)
+{
+	struct pdp_t *pdpa = gsn->pdpa;
 	unsigned int i;
 
 	/* this is O(n) but we don't have (nor want) another hash... */
@@ -231,6 +270,7 @@ int pdp_tidhash(uint64_t tid)
 
 int pdp_tidset(struct pdp_t *pdp, uint64_t tid)
 {
+	struct pdp_t **hashtid = pdp->gsn->hashtid;
 	int hash = pdp_tidhash(tid);
 	struct pdp_t *pdp2;
 	struct pdp_t *pdp_prev = NULL;
@@ -249,6 +289,7 @@ int pdp_tidset(struct pdp_t *pdp, uint64_t tid)
 
 int pdp_tiddel(struct pdp_t *pdp)
 {
+	struct pdp_t **hashtid = pdp->gsn->hashtid;
 	int hash = pdp_tidhash(pdp->tid);
 	struct pdp_t *pdp2;
 	struct pdp_t *pdp_prev = NULL;
@@ -270,6 +311,12 @@ int pdp_tiddel(struct pdp_t *pdp)
 
 int pdp_tidget(struct pdp_t **pdp, uint64_t tid)
 {
+	return gtp_pdp_tidget(g_gsn, pdp, tid);
+}
+
+int gtp_pdp_tidget(struct gsn_t *gsn, struct pdp_t **pdp, uint64_t tid)
+{
+	struct pdp_t **hashtid = gsn->hashtid;
 	int hash = pdp_tidhash(tid);
 	struct pdp_t *pdp2;
 	DEBUGP(DLGTP, "Begin pdp_tidget tid = %"PRIx64"\n", tid);
@@ -286,82 +333,15 @@ int pdp_tidget(struct pdp_t **pdp, uint64_t tid)
 
 int pdp_getimsi(struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi)
 {
-	return pdp_tidget(pdp,
-			  (imsi & 0x0fffffffffffffffull) +
-			  ((uint64_t) nsapi << 60));
+	return gtp_pdp_getimsi(g_gsn, pdp, imsi, nsapi);
 }
 
-/*
-int pdp_iphash(void* ipif, struct ul66_t *eua) {
-  /#printf("IPhash %ld\n", lookup(eua->v, eua->l, ipif) % PDP_MAX);#/
-  return (lookup(eua->v, eua->l, ipif) % PDP_MAX);
+int gtp_pdp_getimsi(struct gsn_t *gsn, struct pdp_t **pdp, uint64_t imsi, uint8_t nsapi)
+{
+	return gtp_pdp_tidget(gsn, pdp, pdp_gettid(imsi, nsapi));
 }
 
-int pdp_ipset(struct pdp_t *pdp, void* ipif, struct ul66_t *eua) {
-  int hash;
-  struct pdp_t *pdp2;
-  struct pdp_t *pdp_prev = NULL;
 
-  if (PDP_DEBUG) printf("Begin pdp_ipset %d %d %2x%2x%2x%2x\n",
-			(unsigned) ipif, eua->l,
-			eua->v[2], eua->v[3],
-			eua->v[4], eua->v[5]);
-
-  pdp->ipnext = NULL;
-  pdp->ipif = ipif;
-  pdp->eua.l = eua->l;
-  memcpy(pdp->eua.v, eua->v, eua->l);
-
-  hash = pdp_iphash(pdp->ipif, &pdp->eua);
-
-  for (pdp2 = haship[hash]; pdp2; pdp2 = pdp2->ipnext)
-    pdp_prev = pdp2;
-  if (!pdp_prev)
-    haship[hash] = pdp;
-  else
-    pdp_prev->ipnext = pdp;
-  if (PDP_DEBUG) printf("End pdp_ipset\n");
-  return 0;
-}
-
-int pdp_ipdel(struct pdp_t *pdp) {
-  int hash = pdp_iphash(pdp->ipif, &pdp->eua);
-  struct pdp_t *pdp2;
-  struct pdp_t *pdp_prev = NULL;
-  if (PDP_DEBUG) printf("Begin pdp_ipdel\n");
-  for (pdp2 = haship[hash]; pdp2; pdp2 = pdp2->ipnext) {
-    if (pdp2 == pdp) {
-      if (!pdp_prev)
-	haship[hash] = pdp2->ipnext;
-      else
-	pdp_prev->ipnext = pdp2->ipnext;
-      if (PDP_DEBUG) printf("End pdp_ipdel: PDP found\n");
-      return 0;
-    }
-    pdp_prev = pdp2;
-  }
-  if (PDP_DEBUG) printf("End pdp_ipdel: PDP not found\n");
-  return EOF; /# End of linked list and not found #/
-}
-
-int pdp_ipget(struct pdp_t **pdp, void* ipif, struct ul66_t *eua) {
-  int hash = pdp_iphash(ipif, eua);
-  struct pdp_t *pdp2;
-  /#printf("Begin pdp_ipget %d %d %2x%2x%2x%2x\n", (unsigned)ipif, eua->l,
-    eua->v[2],eua->v[3],eua->v[4],eua->v[5]);#/
-  for (pdp2 = haship[hash]; pdp2; pdp2 = pdp2->ipnext) {
-    if ((pdp2->ipif == ipif) && (pdp2->eua.l == eua->l) &&
-	(memcmp(&pdp2->eua.v, &eua->v, eua->l) == 0)) {
-      *pdp = pdp2;
-      /#printf("End pdp_ipget. Found\n");#/
-      return 0;
-    }
-  }
-  if (PDP_DEBUG) printf("End pdp_ipget Notfound %d %d %2x%2x%2x%2x\n",
-	 (unsigned)ipif, eua->l, eua->v[2],eua->v[3],eua->v[4],eua->v[5]);
-  return EOF; /# End of linked list and not found #/
-}
-*/
 /* Various conversion functions */
 
 uint64_t pdp_gettid(uint64_t imsi, uint8_t nsapi)
@@ -374,3 +354,17 @@ void pdp_set_imsi_nsapi(struct pdp_t *pdp, uint64_t teid)
 	pdp->imsi = teid & 0x0fffffffffffffffull;
 	pdp->nsapi = (teid & 0xf000000000000000ull) >> 60;
 }
+
+/* Count amount of secondary PDP contexts linked to this primary PDP context
+ * (itself included). Must be called on a primary PDP context. */
+unsigned int pdp_count_secondary(const struct pdp_t *pdp)
+{
+	unsigned int n;
+	unsigned int count = 0;
+	OSMO_ASSERT(!pdp->secondary);
+
+	for (n = 0; n < PDP_MAXNSAPI; n++)
+		if (pdp->secondary_tei[n])
+			count++;
+	return count;
+}
